//! Camera motion estimation using optical flow.
//!
//! Requires the `opencv` feature to be enabled.
//! This module is a placeholder for future implementation.

/// Motion estimator using optical flow.
///
/// This is a placeholder for the OpenCV-based motion estimation
/// that will be implemented when the opencv feature is complete.
pub struct MotionEstimator;
