//! Internal modules ported from external libraries.
//!
//! These modules contain code adapted from:
//! - scipy: Distance metrics
//! - filterpy: Kalman filtering
//! - numpy: Array utilities
//! - motmetrics: MOTChallenge evaluation

pub mod filterpy;
pub mod motmetrics;
pub mod numpy;
pub mod scipy;
